#! /usr/bin/perl
#
# fpgg.pl - Feline Picture Gallery Generator
# Copyright (C) 2006-2008 Tommi Saviranta <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

use strict;
use warnings FATAL => 'all';
use diagnostics;

use File::stat;


# tagval includes global variables:
#   TOTAL-PICTURES
#
my %tagval;

$tagval{'TITLE'} = "Yet Another Gallery";
my $main = "index.html";

my $thumb_size = "128x96";
my $view_size = "640x480";

my $thumb_quality = 85;
my $view_quality = 85;

my $thumb_convert = "convert %s -resize %wx%h -quality %q %d";
my $view_convert = $thumb_convert;
my $thumb_postprocess = "exiv2 -d a %f";
my $view_postprocess =
	"exiv2 \"-M del Exif.Photo.MakerNote\" %f; exiv2 -d t %f;";

my $img_basepath = "";

my ($thumb_width, $thumb_height);
my ($view_width, $view_height);
my ($thumb_true_size, $view_true_size);

my %all_files;
my ($list_all_files, $ssh_all_files);

my $cleanup = 0;
my $trust_size = 1;

my ($template_index, $template_view, $template_category);

my @file_id;
my (%file_name, %file_dir, %file_cat, %file_md5);
my (%file_thumb_w, %file_thumb_h);
my %categories;

# coordiantes -> maps
my @maps;
my %maps;

# size, md5, width, height, id, name
my %file_cache;
my @get_info_cache = ();
my $get_info_cache_file = "";

# maps
push(@maps, "google");
$maps{'google-name'} = 'Google Maps';
#
# URL format:
# http://maps.google.com/?ll=60.172583,24.932758&z=17&t=h
#	ll=y,z	coordinates, longitude,latitude in decimal
#	z=x	zoom level
#	t=x	map type
#			h = aerial with labels
#			k = aerial without labels
#			p = terrain
#
$maps{'google-format'} = 'http://maps.google.com/?ll=[lat],[lon]&z=[zoom]&t=h';
$maps{'google-default-zoom'} = 15;
# zoom level North-South ranges around 50 N
$maps{'google-zoom'} = '25=21, 45=20, 90=19, 190=18, 390=17, '
		. '750=16, 1450=15, 3000=14, 6250=13, 12500=12, '
		. '22500=11, 45000=10, 90000=9, 180000=8, '
		. '350000=7, 675000=6, 1250000=5, 3000000=4, '
		. '6500000=3, 11000000=2, 20000000=1, 99999999=0';

push(@maps, "microsoft");
$maps{'microsoft-name'} = 'Virtual Earth';
#
# URL format (version 2?):
# http://maps.live.com/default.aspx?v=2&cp=60.172524~24.932951&style=h&lvl=18
# 	v=2	version info, required
#	cp=y~x	coordinate pair, latitude~longitude in decimal
#	lvl=x	zoom level
#	style=x	map style
#			r = road
#			h = aerial
#
$maps{'microsoft-format'} = 'http://maps.live.com/default.aspx?v=2'
		. '&cp=[lat]~[lon]&lvl=[zoom]&style=h';
$maps{'microsoft-default-zoom'} = 15;
# zoom level North-South ranges around 50 N
$maps{'microsoft-zoom'} = '80=19, 160=18, 360=17, 85=16, 1500=15, 3000=14, '
		. '6000=13, 15000=12, 25000=11, 45000=10, 90000=9, 180000=8, '
		. '380000=7, 700000=6, 1600000=5, 3600000=4, 6500000=3, '
		. '13300000=2, 99999999=1';

push(@maps, "yahoo");
$maps{'yahoo-name'} = 'Yahoo! Maps';
#
# URL format:
# http://maps.yahoo.com/#mvt=h&lat=52.518545&lon=13.376154&zoom=17
# 	mvt=x	map type
#			s = satellite
#			m = map
#			h = hybrid
#	lat=x	latitude in decimal
#	lon=x	longiture in decimal
#	zoom=x	zoom level
#
$maps{'yahoo-format'} = 'http://maps.yahoo.com/#mvt=h&"
		. "lat=[lat]&lon=[lon]&zoom=[zoom]';
$maps{'yahoo-default-zoom'} = 15;
# zoom level North-South ranges around 50 N	FIXME
$maps{'yahoo-zoom'} = '80=19, 160=18, 360=17, 85=16, 1500=15, 3000=14, '
		. '6000=13, 15000=12, 25000=11, 45000=10, 90000=9, 180000=8, '
		. '380000=7, 700000=6, 1600000=5, 3600000=4, 6500000=3, '
		. '13300000=2, 99999999=1';



=item print_usage
Print usage and quit
=cut
sub print_usage() {
	print <<EOF;
Usage: fpgg.pl [OPTION ...] dir ...

  --title x		Gallery title

  --view-size		Picture dimensions in view-page (WIDTHxHEIGHT)
  --view-quality	Picture quality (0-100, 100 = best)
  --view-convert CMD	Use given command to convert original into view
			pictures. "%s", "%d", "%w", "%h", and "%q", without
			quotes, are expanded into source file, destination
			file, width, height, and quality, respectively. Do not
			place quotes around %s or %d, these are added
			automatically. Default value uses "convert" for
			conversion.
  --view-postprocess CMD	Use given command to post-process view
 			pictures. E.g. can be used to strip unwanted EXIF-data
			from the picture to reduce file size. Multiple commands
			can be given by separating them with a semicolon. All
			occurances of "%s" (without quotes) are expanded with
			the filename. Default value uses "exiv2" to strip
			MakerNote and thumbnail data. Give empty string to
			disable.

  --thumb-size		Thumbnail dimensions (WIDTHxHEIGHT)
  --thumb-quality	Thumbnail quality (0-100, 100 = best)
  --thumb-convert CMD	Like --view-convert, but for thumbnails.
  --thumb-postprocess CMD	Like --view-postprocess, but for thumbnails.
			Unlike --view-postprocess, default value for this
			strips all metadata.

  --img-basepath URI	Define hardcoded basepath for images
			(Allows to have images on an external server)

  --list-all-files	List all files that are needed for this gallery.
  --ssh-all-files HOST	Like --list-all-files, but prints out a full command to
  			copy the files to remote host using ssh and tar.
EOF
	exit(0);
}



=item parse_options
Parse command line options
=cut
sub parse_options() {
	while (defined($ARGV[0]) && ($ARGV[0] =~ /^--./)) {
		if ($ARGV[0] eq "--title") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "title missing\n";
				exit(1);
			}
			$tagval{'TITLE'} = $ARGV[0];
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--view-size") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "view-size missing\n";
				exit(1);
			}
			$view_size = $ARGV[0];
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--thumb-size") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "thumb-size missing\n";
				exit(1);
			}
			$thumb_size = $ARGV[0];
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--view-quality") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "view-quality missing\n";
				exit(1);
			}
			$view_quality = $ARGV[0];
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--thumb-quality") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "thumb-quality missing\n";
				exit(1);
			}
			$thumb_quality = $ARGV[0];
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--img-basepath") {
			shift @ARGV;
			if (! defined($ARGV[0])) {
				print "img-basepath missing\n";
				exit(1);
			}
			$img_basepath = $ARGV[0];
			$img_basepath .= "/" unless ($img_basepath =~ /\/$/);
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--list-all-files") {
			$list_all_files = 1;
			shift @ARGV;
		}

		elsif ($ARGV[0] eq "--ssh-all-files") {
			shift @ARGV;
			$ssh_all_files = shift @ARGV;
		}

		elsif ($ARGV[0] eq "--help") {
			&print_usage();
			exit(0);
		}

		else {
			print "Unknown option: $ARGV[0]\n";
			exit(1);
		}
	}
}



=item read_file
Read file

0: filename
1: failure is fatal
r: (lines without linefeed or undef)
=cut
sub read_file($$) {
	my ($file, $fatal) = @_;
	my $r = open(FILE, "< $file");
	if (! defined($r)) {
		die "$!" if ($fatal == 1);
		return undef;
	}
	my @t = <FILE>;
	chomp @t;
	return @t;
}



=item get_tagval_thumbs
Get tagval for ALL-THUMBS
=cut
sub get_tagval_all_thumbs() {
	my $t = "";
	foreach my $num (@file_id) {
		$t .= &create_thumb_link($num) . "\n";
	}
	$tagval{'ALL-THUMBS'} = $t;
}



=item get_tagval_all_categories
Get tagval for ALL-CATEGORIES
=cut
sub get_tagval_all_categories() {
	my @cats = keys %categories;
	if ($#cats <= 0) {
		$tagval{'ALL-CATEGORIES'} = "";
		return;
	}

	my $r = "<ul>\n";
	foreach my $cat (sort {uc($a) cmp uc($b)} keys %categories) {
		my $catfile = &make_cat_file($cat);
		$r .= "\t<li><a href=\"$catfile\">$cat</a></li>\n";
	}
	$r .= "</ul>\n";
	$tagval{'ALL-CATEGORIES'} = $r;
}


=item get_tags_for_all_files
Update %all_files based on template and tags

0: data
=cut
sub get_tags_for_all_files() {
	$_ = shift;
	foreach my $tag ("PICTURE-LINK-ORIGINAL",
			"PICTURE-VIEW-LINK-ORIG") {
		next unless (/<\*?TAG:\Q$tag\E[:>]/);
		my $file = $tagval{"$tag-URL"};
		$file =~ s/\Q$img_basepath\E//;
		$all_files{$file} = 1;
	}
}



=item replace_simple_tags
Replace simple tags in data

0: data
r: data
=cut
sub replace_simple_tags($) {
	$_ = shift;

	foreach my $tag ("TITLE", "TOTAL-PICTURES", "CREATED",
			"ALL-THUMBS", "ALL-CATEGORIES",
			"CATEGORY", "CATEGORY-THUMBS",
			"VIEW-OPTIONS",
			"PICTURE-FULLPATH", "PICTURE-FILE", "PICTURE-PATH",
			"PICTURE-TITLE", "PICTURE-DESCRIPTION",
			"PICTURE-CATEGORIES", "PICTURE-I",
			"PICTURE-VIEW", "PICTURE-VIEW-LINK-ORIG",
			"PICTURE-VIEW-LINK-NEXT",
			"ALL-THUMBS", "ALL-CATEGORIES") {
		my $val = $tagval{$tag};
		if (! defined($val)) {
			s/(.*\n).*?<\*TAG:\Q$tag\E>.*?\n/$1/gs;
		} else {
			s/<\*?TAG:\Q$tag\E>/$val/gs;
		}
	}
	return $_;
}



=item get_map_level_str
Convert map detai level ID to string

0: level id
1: level 1 string
2: level 2 string
r: level string
=cut
sub get_map_level_str($$$) {
	my ($id, $lev1, $lev2) = @_;

	if ($id == 1) {
		return defined($lev1) ? $lev1 : "Location";
	} elsif ($id == 2) {
		return defined($lev2) ? $lev2 : "Approximate location";
	}
	return "Location";
}



=item replace_opt_tags
Replace tags with options

0: data
r: data
=cut
sub replace_opt_tags($) {
	$_ = shift;

	# replace single option tags
	foreach my $tag ("VIEW-LINK-PREV", "VIEW-LINK-NEXT",
			"LINK-INDEX", "PICTURE-LINK-ORIGINAL") {
		my $replace;
		# replace tags with option
		foreach my $macro (/(<TAG:$tag:[^>]+>)/gs) {
			my ($label) = $macro =~ /.*:(.*)>/s;
			$replace = sprintf("<a href=\"%s\">%s</a>",
					$tagval{"$tag-URL"}, $label);
			s/\Q$macro\E/$replace/s;
		}

		# replace remaining tags
		my $url = $tagval{"$tag-URL"};
		if (defined($url)) {
			$replace = sprintf("<a href=\"%s\">%s</a>",
					$url, $tagval{"$tag-LABEL"});
		} else {
			$replace = $tagval{"$tag-LABEL"};
		}
		s/<TAG:\Q$tag\E>/$replace/gs;
	}

	# replace multi-option tags
	foreach my $macro (/(<\*?TAG:PICTURE-MAP[^>]*>)/gs) {
		my @t = split(/[<>:]/, $macro);
		my $links;

		# multi-service maps
		$links = $tagval{"PICTURE-MAP-LINKS"};
		if (defined($links)) {
			my $level = $tagval{"PICTURE-MAP-LINKS-DETAIL"};
			my $label = &get_map_level_str($level, $t[3], $t[4]);
			s/\Q$macro\E/$label: $links/s;
			next;
		}

		# absolute url
		$links = $tagval{'PICTURE-MAP-URL'};
		if (defined($links)) {
			my $level = $tagval{'PICTURE-MAP-LEVEL'};
			my $label = &get_map_level_str($level, $t[3], $t[4]);
			my $val = sprintf("<a href=\"%s\">%s</a>\n",
					$tagval{'PICTURE-MAP-URL'},
					$label);
			s/\Q$macro\E/$val/s;
			next;
		}

		# no service/url

		if ($t[1] =~ /^\*/) {
			s/(.*\n).*?<\*TAG:PICTURE-MAP[^>]*>.*?\n/$1/gs;
		} else {
			s/\Q$macro\E//s;
		}
	}
	
	return $_;
}



=item replace_tags
Replace tags

0: data
r: data
=cut
sub replace_tags($) {
	my ($t) = @_;
	&get_tags_for_all_files($t);
	$t = &replace_simple_tags($t);
	$t = &replace_opt_tags($t);
	return $t;
}



=item get_size
Get file size

0: path
r: size
=cut
sub get_size($) {
	my ($file) = @_;
	my $sb = stat($file) or die "stat($file): $!";
	return $sb->size;
}



=item get_md5
Get file MD5

0: file
1: md5sum
=cut
sub get_md5($) {
	my ($file) = @_;
	open(MPIPE, "md5sum $file |") or die $@;
	my $line = <MPIPE>;
	close(MPIPE);
	my ($md5) = $line =~ /(.*?) /;
	return $md5;
}



=cut cache_get_file_id
Get file ID from cache

0: picture file
r: ID
=cut
sub cache_get_file_id($) {
	my ($file) = @_;
	my @r = &cache_get($file);
	return $r[4];
}



=item cache_get
Get picture info from cache

0: picture file
=cut
sub cache_get($) {
	my ($file) = @_;
	my $line = $file_cache{$file};
	return () if (! defined($line));

	my ($size, $md5, $width, $height, $id) = $line =~
			/(\d+)\t([0-9a-f]+)\t(\d+)\t(\d+)\t(\d+)/i;
	return ($size, $md5, $width, $height, $id);
}



=item cache_set
Write file info into cache

0: file
1: file size
2: md5
3: width
4: height
5: number
=cut
sub cache_set($$$$$$) {
	my ($file, $size, $md5, $width, $height, $num) = @_;
	$file_cache{$file} = "$size\t$md5\t$width\t$height\t$num";
}



=item get_file_info
Get image info

0: directory
1: file
2: global id
r: (id, width, height, need_rebuild)
=cut
sub get_file_info($$$) {
	my ($dir, $file, $gid) = @_;
	my ($size, $md5, $width, $height, $id) = &cache_get("$dir/$file");
	my $rebuild;

	if (defined($id)) {
		if ($trust_size == 1) {
			my $tsize = &get_size("$dir/$file");
			$rebuild = ($tsize == $size) ? 0 : 1;
		} else {
			my $tmd5 = &get_md5("$dir/$file");
			$rebuild = ($tmd5 eq $md5) ? 0 : 1;
		}
	}
	if (! defined($id) || $rebuild == 1) {
		$size = &get_size("$dir/$file");
		print ".";
		$md5 = &get_md5("$dir/$file");
		print ".";
		($width, $height) = &get_dimensions("$dir/$file");
		print ".";
		$id = $gid;
		&cache_set("$dir/$file", $size, $md5, $width, $height, $id);
		$rebuild = 1;
	}
	return ($id, $width, $height, $rebuild);
}



=item get_new_id
Get new ID
TODO: what is this?-)

r: ID
=cut
sub get_new_id() {
	my $big_id = -1;
	while (my ($key, $val) = each %file_cache) {
		my $id = &cache_get_file_id($key);
		$big_id = $id if ($id > $big_id);
	}
	return $big_id + 1;
}



=item cache_load
Load picture information cache

0: file
=cut
sub cache_load($) {
	my ($file) = @_;
	open(CFILE, "< $file") or return -1;
	while(<CFILE>) {
		my ($val, $key) = /(.*)\t(.*)/;
		$file_cache{$key} = $val;
	}
	close(CFILE);
}



=item cache_save
Save cached information

0: file
=cut
sub cache_save($) {
	my ($file) = @_;
	open(CFILE, "> $file") or die $@;
	while (my ($key, $val) = each %file_cache) {
		print CFILE "$val\t$key\n";
	}
	close(CFILE);
}



=item split_categories
Split string into categories

0: string
r: (categories)
=cut
sub split_categories($) {
	my ($catstr) = @_;
	return () if (! defined($catstr));
	my @cats = split(/,\s*/, $catstr);
	return @cats;
}



=item add_to_cats
Add image to categories

0: filename
1: directory
2: number
3: categories
4: thumb width
5: thumb height
=cut
sub add_to_cats($$$$$$) {
	my ($file, $dir, $num, $cats, $width, $height) = @_;
	$categories{$_} .= "$num " foreach (&split_categories($cats));

	die "sanity check" if (defined($file_name{$num}));
	$file_name{$num} = $file;

	die "sanity check" if (defined($file_dir{$num}));
	$file_dir{$num} = $dir;
	
	die "sanity check" if (defined($file_cat{$num}));
	if (defined($cats)) {
		$file_cat{$num} = "$cats, All";
	} else {
		$file_cat{$num} = "All";
	}
	
	die "sanity check" if (defined($file_thumb_w{$num}));
	$file_thumb_w{$num} = $width;

	die "sanity check" if (defined($file_thumb_h{$num}));
	$file_thumb_h{$num} = $height;

	push(@file_id, $num);
}



=item create_thumb_link
Create thumbnail link

0: image number
r: string
=cut
sub create_thumb_link($) {
	my ($num) = @_;

	my $dir = $file_dir{$num};
	my $file = $file_name{$num};
	my $thumb = &make_thumb_name($num, $dir, $file);

	my $alt = &get_info($dir, $file, "title");
	$alt = "$dir/$file" unless defined($alt);

	$all_files{$thumb} = 1;
	return sprintf("<a href=\"%s\"><img src=\"%s\"\n" .
			"\tclass=\"%s\" width=\"%d\" height=\"%d\"\n" .
			"\talt=\"%s\" /></a>",
			&make_view_html_name($num),
			$img_basepath . $thumb,
			($file_thumb_w{$num} > $file_thumb_h{$num}) ? "v" : "h",
			$file_thumb_w{$num}, $file_thumb_h{$num},
			$alt);
}



=item build_category
Build category page

0: category
=cut
sub build_category($) {
	my ($cat) = @_;

	if (! defined($template_category)) {
		$template_category =
			join("\n", &read_file("category.template", 1));
	}

	$tagval{'CATEGORY'} = $cat;

	my $t = "";
	foreach my $num (split(/ /, $categories{$cat})) {
		$t .= &create_thumb_link($num) . "\n";
	}
	$tagval{'CATEGORY-THUMBS'} = $t;

	my $out_file = &make_cat_file($cat);
	my $page = $template_category;

	$page = &replace_tags($page);

	$all_files{$out_file} = 1;
	open(FILE, "> $out_file") or die $@;
	print FILE $page;
	close(FILE);
}



=item build_cats
Build category pages
=cut
sub build_cats() {
	&build_category($_) foreach (keys %categories);
}



=item get_info
Get info of picture

0: directory
1: file
2: field
r: info
=cut
sub get_info($$$) {
	my ($dir, $file, $field) = @_;
	my $phase = 0;
	my $ret;
	if ($get_info_cache_file ne "$dir/info") {
		@get_info_cache = &read_file("$dir/info", 0);
		return undef unless defined($get_info_cache[0]);
		$get_info_cache_file = "$dir/info";
	}

	foreach (@get_info_cache) {
		if (/^file: /) {
			($phase != 0) && return $ret;
			(/^file: $file/) || next;
			$phase = 1;
			next;
		}
		if ($phase == 1 && /^$field:/) {
			($ret) = /.*?: (.*)/;
			$phase = 2;
			next;
		}
		if ($phase == 2 && (/^$/ || /^[^ ]+:/)) {
			return $ret;
		}
		$ret .= "\n$_" if ($phase == 2);
	}
	return $ret;
}



=item get_dimensions
Get image dimensions

0: file
r: (width, height)
=cut
sub get_dimensions($$) {
	my ($file) = @_;
	open(SPIPE, "identify \"$file\" |") or die "pipe failed!";
	my $out = <SPIPE>;
	close(SPIPE);
	my ($width, $height) = $out =~ /.* (\d+)x(\d+)/;
	return ($width, $height);
}



=item scale_size
Scale dimensions

0: original width
1: original height
2: mode
r: (width, height)
=cut
sub scale_size($$$) {
	my ($width, $height, $mode) = @_;
	my ($swidth, $sheight);
	my $div;

	if ($mode eq "thumb") {
		$swidth = $thumb_width;
		$sheight = $thumb_height;
	} else {
		$swidth = $view_width;
		$sheight = $view_height;
	}

	if ($width > $height) {
		$div = $width / $swidth;
	} else {
		my $t = $swidth;
		$swidth = $sheight;
		$sheight = $t;
		$div = $height / $sheight;
	}

	return ($width / $div, $height / $div);
}



=item convert
Convert picture.

0: command
1: source
2: destination
3: width
4: height
5: quality
=cut
sub convert($$$$$$) {
	my ($cmd, $src, $dest, $width, $height, $quality) = @_;
	$cmd =~ s/%s/\"$src\"/g;
	$cmd =~ s/%d/\"$dest\"/g;
	$cmd =~ s/%w/$width/g;
	$cmd =~ s/%h/$height/g;
	$cmd =~ s/%q/$quality/g;
	system($cmd);
}



=item postprocess
Postprocess picture.

0: command
1: file
=cut
sub postprocess($$) {
	my ($cmd, $file) = @_;

	if (defined($cmd) && $cmd ne "") {
		$cmd =~ s/\%f/\"$file\"/g;
		system($cmd);
		print ".";
	}
}



=item make_filename
Make filename

0: directory
1: number
2: extension
r: filename
=cut
sub make_filename($$$) {
	my ($dir, $num, $ext) = @_;
	my ($n) = "00000000$num" =~ /(........)$/;
	$_ = "$dir$n$ext";
	s,//+,/,g;
	return $_;
}



=item make_cat_file
Make category filename

0: category
r: filename
=cut
sub make_cat_file($) {
	my ($cat) = @_;
	$cat =~ s/ /_/g;
	$cat =~ tr/A-Z/a-z/;
	$cat =~ s/[\[\]()]//g;
	$cat =~ s/&([a-z])[a-z]+;/$1/ig;
	return "cat_$cat.html";
}



=item make_view_pic_name
Make view picture filename

0: number
1: directory
2: filename
r: filepath
=cut
sub make_view_pic_name($$$) {
	my ($num, $dir, $file) = @_;
	return &make_filename("${dir}_r/view_", $num, "_$file");
}



=item make_view_html_name
Make view.html filename

0: picture number
r: filename
=cut
sub make_view_html_name($) {
	my ($num) = @_;
	return &make_filename("", $num, ".html");
}



=item make_thumb_name
Make thumbnail picture filename

0: number
1: directory
2: filename
r: filename
=cut
sub make_thumb_name($$$) {
	my ($num, $dir, $file) = @_;
	return &make_filename("${dir}_r/thumb_", $num, "_$file");
}



=item make_view_view
Create view picture

0: source
1: destination
=cut
sub make_view_pic($$) {
	my ($src, $dest) = @_;
	&convert($view_convert, $src, $dest,
			$view_true_size, $view_true_size,
			$view_quality);
	print ".";

	&postprocess($view_postprocess, $dest);
	print ".";
}



=item add_view
Add view to be created

0: number
1: directory
2: filename
3: rebuilding needed
=cut
sub add_view($$$$) {
	my ($num, $dir, $file, $rebuild) = @_;

	my ($id, $width, $height) = &get_file_info($dir, $file, $num);
	die "sanity check" if ($id != $num);

	my $real_name = "$dir/$file";
	my $view_name = &make_view_pic_name($num, $dir, $file);

	&make_view_pic($real_name, $view_name) if ($rebuild == 1);

	my $cats = &get_info($dir, $file, "category");
	my ($twidth, $theight) = &scale_size($width, $height, "thumb");
	&add_to_cats($file, $dir, $num, $cats, $twidth, $theight);
}



=item build_views
Create views
=cut
sub build_views() {
	foreach my $num (@file_id) {
		&build_view($num);
	}
}



=item parse_text
Parse text

0: text
r: text
=cut
sub parse_text($) {
	my ($text) = @_;
	my $fix;
	return undef if (! defined($text));
	foreach my $macro ($text =~ /\[\[\s*?(.*?)\s*?\]\]/gs) {
		my @param = split(/\]\s+\[/, $macro);
		if ($#param != 1) {
			print "bad macro: \[\[$macro\]\]: bad parameters\n";
			return $text;
		}
		my $id = &cache_get_file_id($param[1]);
		if (! defined($id)) {
			print "bad marco: \[\[$macro\]\]: file not found\n";
			return $text;
		}

		my $view = &make_view_html_name($id);
		my $repl = "<a href=\"$view\">$param[0]</a>";
		$text =~ s/\[\[\Q$macro\E\]\]/$repl/s;
	}
	return $text;
}



=item split_coordinate
Split coordiante into lat/lon.

0: string
r: lat/lon
=cut
sub split_coordinate($) {
	my ($str) = @_;
	my $old_str = $str;
	my $mul = 0;

	my ($d) = $str =~ /(.)$/;
	$str =~ s/\s*[NSEW]$//i;

	my $deg = 0;
	my @parts = split(/[ \-_\/]/, $str);
	my $div = 1;
	foreach my $part (@parts) {
		$deg += $part / $div;
		$div *= 60;
	}

	$deg = -$deg if ($d =~ /[SW]/i);
	return $deg;
}



=item split_coordinates
Split coordinate string into latitude and longitude

0: string
r: (latitude, longitude, detail)
=cut
sub split_coordinates($$) {
	my ($str) = @_;
	my ($lat, $lon, $detail);

	if ($str =~ /,/) {
		($lat, $lon, $detail) = split(/,\s*/, $str);

		# assume lat, lon, but swap if not so
		if (($lon =~ /[NS]$/i) || ($lat =~ /[EW]$/i)) {
			my $t = $lat;
			$lat = $lon;
			$lon = $t;
		}

		$lat = &split_coordinate($lat);
		$lon = &split_coordinate($lon);
		$detail =~ s/\s*M//i;
	} else {
		($lat, $lon, $detail) = split(/\s+/, $str);
	}

	return ($lat, $lon, $detail);
}



=item coord_to_urls
0: latitude
1: longitude
2: detail

r: (urls, level)
=cut
sub coord_to_urls($$$) {
	my ($lat, $lon, $detail) = @_;
	my $out = "";

	foreach my $map (@maps) {
		my $zoom = $maps{"$map-default-zoom"};
		foreach my $t_zoom (split(/,\s*/, $maps{"$map-zoom"})) {
			my ($d, $z) = split(/=/, $t_zoom, 2);
			if ($detail <= $d) {
				$zoom = $z;
				last;
			}
		}

		my $uri = $maps{"$map-format"};
		$uri =~ s/\[lat\]/$lat/g;
		$uri =~ s/\[lon\]/$lon/g;
		$uri =~ s/\[zoom\]/$zoom/g;

		$out .= sprintf(",\n\t<a href=\"%s\">%s</a>", $uri,
				$maps{"$map-name"});
	}

	$out =~ s/^,//;
	return ($out, $detail <= 1000 ? 1 : 2);
}



=item build_view
Create view

0: id
=cut
sub build_view($) {
	my ($num) = @_;

	my $t;
	my $dir = $file_dir{$num};
	my $file = $file_name{$num};
	my ($id, $width, $height, $dummy) = &get_file_info($dir, $file, $num);
	my ($vwidth, $vheight) = &scale_size($width, $height, "view");

	my $real_name = "$dir/$file";
	my $view_name = &make_view_pic_name($num, $dir, $file);

	my $orig_file = &get_info($dir, $file, "orig");
	$orig_file = $real_name unless defined($orig_file);

	# title
	$tagval{'PICTURE-TITLE'} = &get_info($dir, $file, "title");

	if (! defined($tagval{'PICTURE-TITLE'})) {
		$tagval{'PICTURE-TITLE'} = $real_name;
	}

	my $out_file = &make_view_html_name($num);
	print "$real_name -> $out_file\n";

	# description
	my $t_text = &get_info($dir, $file, "text");
	$t_text = &parse_text($t_text);
	$tagval{'PICTURE-DESCRIPTION'} = $t_text;

	# categories
	undef $tagval{'PICTURE-CATEGORIES'};
	my $cats = &get_info($dir, $file, "category");
	my @cats = &split_categories($cats);
	if ($#cats >= 0) {
		my $t = "";
		foreach (sort {uc($a) cmp uc($b)} @cats) {
			my $cat = &make_cat_file($_);
			$t .= "<a href=\"$cat\">$_</a>\n";
		}
		$tagval{'PICTURE-CATEGORIES'} = $t;
	}

	# map, absolute url
	undef $tagval{'PICTURE-MAP-URL'};
	my $t_map = &get_info($dir, $file, "map");
	$tagval{'PICTURE-MAP-LEVEL'} = 0;
	if (defined($t_map)) {
		$tagval{'PICTURE-MAP-URL'} = $t_map;
		$tagval{'PICTURE-MAP-LEVEL'} = 1;
	} else {
		my $t_map_ap = &get_info($dir, $file, "map-ap");
		if (defined($t_map_ap)) {
			$tagval{'PICTURE-MAP-URL'} = $t_map_ap;
			$tagval{'PICTURE-MAP-LEVEL'} = 2;
		}
	}

	# map, any service
	undef $tagval{'PICTURE-MAP-LINKS'};
	my $t_coord = &get_info($dir, $file, "coords");
	my ($coord_lat, $coord_lon, $coord_detail);
	if (defined($t_coord)) {
		($coord_lat, $coord_lon, $coord_detail) =
				&split_coordinates($t_coord);
	}
	if (defined($coord_lat) && defined($coord_lon)) {
		($tagval{'PICTURE-MAP-LINKS'},
				$tagval{'PICTURE-MAP-LINKS-DETAIL'})
			= &coord_to_urls($coord_lat, $coord_lon, $coord_detail);
	}

	# view
	$all_files{$view_name} = 1;
	$tagval{'PICTURE-VIEW'} =
		sprintf("<img src=\"%s\" width=\"%d\" height=\"%d\"\n" .
				"\talt=\"%s\" />",
			"$img_basepath$view_name",
			$vwidth, $vheight,
			$tagval{'PICTURE-TITLE'});
	$tagval{'PICTURE-VIEW-LINK-ORIG'} =
		"<a href=\"$img_basepath$orig_file\">\n" .
		$tagval{'PICTURE-VIEW'} . "</a>";
	$tagval{'PICTURE-VIEW-LINK-ORIG-URL'} = "$img_basepath$orig_file";
	if ($num + 1 != $tagval{'TOTAL-PICTURES'}) {
		$t = &make_view_html_name($num + 1);
	} else {
		$t = $tagval{'LINK-INDEX-URL'};
	}
	$tagval{'PICTURE-VIEW-LINK-NEXT'} =
		"<a href=\"$t\">\n" . $tagval{'PICTURE-VIEW'} . "</a>";

	# prev/next
	undef $tagval{'VIEW-LINK-PREV-URL'};
	$tagval{'VIEW-LINK-PREV-LABEL'} = "&lt;&lt;&lt;";
	if ($num != 0) {
		$tagval{'VIEW-LINK-PREV-URL'} = &make_view_html_name($num - 1);
	}
	undef $tagval{'VIEW-LINK-NEXT-URL'};
	$tagval{'VIEW-LINK-NEXT-LABEL'} = "&gt;&gt;&gt;";
	if ($num + 1 != $tagval{'TOTAL-PICTURES'}) {
		$tagval{'VIEW-LINK-NEXT-URL'} = &make_view_html_name($num + 1);
	}

	# original
	$tagval{'PICTURE-LINK-ORIGINAL-URL'} = "$img_basepath$orig_file";
	$tagval{'PICTURE-LINK-ORIGINAL-LABEL'} = "Original";

	# misc
	$tagval{'PICTURE-I'} = $num + 1;
	$tagval{'PICTURE-FULLPATH'} = "$img_basepath$real_name";
	$tagval{'PICTURE-PATH'} = "$img_basepath";
	$tagval{'PICTURE-FILE'} = "$real_name";


	# create page
	my $page = $template_view;

	$page = &replace_tags($page) . "\n";

	$all_files{$out_file} = 1;
	open(FILE, "> $out_file") or die $@;
	print FILE $page;
	close(FILE);
}



=item make_thumb_pic
Make thumbnail picture

0: source
1: destination
=cut
sub make_thumb_pic($$) {
	my ($src, $dest) = @_;
	&convert($thumb_convert, $src, $dest,
			$thumb_true_size, $thumb_true_size,
			$thumb_quality);
	print ".";

	&postprocess($thumb_postprocess, $dest);
	print ".";
}



=item add_pic
Add picture to be processed

0: picture number
1: directory
2: filename
r: ID
=cut
sub add_pic($$$) {
	my ($num, $dir, $file) = @_;

	print "$dir/$file .";
	# use ID, it will be globally correct
	my ($id, $width, $height, $rebuild) = &get_file_info($dir, $file, $num);
	if ($rebuild == 1) {
		my $thumb = &make_thumb_name($id, $dir, $file);
		&make_thumb_pic("$dir/$file", $thumb);
	}

	&add_view($id, $dir, $file, $rebuild);
	print "\n";
	return $id;
}



=item add_dir
Add new directory to be scanned

0: first picture number to be assigned
1: directory
=cut
sub add_dir($$) {
	my ($num, $dir) = @_;
	print "dir: $dir\n";
	if ($cleanup == 1) {
		system("test -d \"${dir}_r/\" && rm -rf \"${dir}_r\"");
	}
	system("test -d \"${dir}_r/\" || mkdir \"${dir}_r\"");
	open(DPIPE, "find \"$dir/\" -iname '*.jp*g' | sort |") or die $@;
	while (<DPIPE>) {
		my ($file) = /.*\/(.*)/;
		$num = &add_pic($num, $dir, $file) + 1;
	}
	close(DPIPE);
}



=cut get_total_count
Get total count of pictures

@: directories
r: number of jpeg-pictures (with correct suffix)
=cut
sub get_total_count() {
	my $r;
	open(CPIPE, "find @_ -iname '*.jp*g' | wc -l |") or die $@;
	$r = <CPIPE>;
	close(CPIPE);
	chomp $r;
	return $r;
}



=item max
Get maximum of two

0: a
1: b
r: greater of a and b
=cut
sub max($$) {
	if ($_[0] > $_[1]) {
		return $_[0];
	} else {
		return $_[1];
	}
}



=item build_index
Build index page
=cut
sub build_index() {
	my $out_file = $main;
	my $page = $template_index;

	$page = &replace_tags($page);

	$all_files{$out_file} = 1;
	open(FILE, "> $out_file") or die $@;
	print FILE $page;
	close(FILE);
}



=item main
Main function

@: directories
=cut
sub main(@) {
	($thumb_width, $thumb_height) = $thumb_size =~ /(\d+)x(\d+)/;
	$thumb_true_size = &max($thumb_width, $thumb_height);
	($view_width, $view_height) = $view_size =~ /(\d+)x(\d+)/;
	$view_true_size = &max($view_width, $view_height);

	$tagval{'TOTAL-PICTURES'} = &get_total_count(@ARGV);
	$tagval{'LINK-INDEX-URL'} = $main; # XXX
	$tagval{'LINK-INDEX-LABEL'} = "Back to index";
	$tagval{'CREATED'} = localtime();
	my $id;

	if ($tagval{'TOTAL-PICTURES'} == 0) {
		print "Nothing to do.\n";
		return;
	}

	print "Building pictures...\n";
	&cache_load("info_cache") if ($cleanup == 0);
	foreach (@ARGV) {
		$_ =~ s,/$,,;
		print "$_\n";
		$id = &get_new_id();
		&add_dir($id, $_);
	}
	&cache_save("info_cache");

	&get_tagval_all_thumbs();
	&get_tagval_all_categories();

	print "\nBuilding HTML...\n";
	&build_views();
	&build_cats();
	&build_index();

	if (defined($list_all_files)) {
		print "\nAll files:";
		print $_ foreach (keys %all_files);
	}

	if (defined($ssh_all_files)) {
		my ($user, $dir);
		($user, $ssh_all_files) = split(/\@/, $ssh_all_files, 2)
				if ($ssh_all_files =~ /\@/);
		($ssh_all_files, $dir) = split(/:/, $ssh_all_files, 2)
				if ($ssh_all_files =~ /:/);

		print "tar -czhf -";
		print " $_" foreach (keys %all_files);
		print " | ssh";
		print " -l $user" if (defined($user));
		print " $ssh_all_files '";
		print "cd \"$dir\" && " if (defined($dir));
		print "tar -xzvvf -'\n";
	}
}

if ($#ARGV == -1) {
	&print_usage();
	exit(0);
} else {
	&parse_options();
}

$template_view = join("\n", &read_file("view.template", 1));
$template_index = join("\n", &read_file("index.template", 1));
# category template is read on demand

&main(@ARGV);
